import logging
import socket
import sys
import os
from threading import Thread
import argparse
import json
import re
import requests
import time
from utils import _select_folder, _make_imap_client, logout_client

threads = {}
client_count = 0
HOST = '0.0.0.0'
MAX_LENGTH = 4096
PORT = 10010
idle_timeout = 30
SSL_VERIFY = False
cs_host = 'localhost'
LOG_FILE_PATH = '/var/log/cyops/cyops-integrations/imap/listener.log'
logging.basicConfig(filename=LOG_FILE_PATH, level=logging.INFO,
                    format='%(asctime)s %(levelname)s %(module)s %(funcName)s(): %(message)s')
logger = logging.getLogger(__name__)
serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

def find_byte_value(data, find_value, key = None):
    for value in data:
        str_value = value.decode("utf-8").strip("\\")
        if str_value == find_value:
            return True
    return False

def find_unread_mail(responses):
    try:
        for response_flag in responses:
            command = response_flag[1].decode("utf-8")
            if command == 'FETCH' and response_flag[0]>0:
                fetch_result = response_flag[2][3];
                if not find_byte_value(fetch_result,'Seen'):
                    return True
        return False
    except Exception as e:
        logger.exception("No match found in response by idle check %s" % e)
        return False

def get_auth_token(username, password):
    endpoint = 'https://' + cs_host + '/auth/authenticate'
    headers = {'Content-Type': 'application/json'}
    data = {'credentials': {'loginid': username, 'password': password, 'token': ''}}
    ret = requests.post(endpoint, data=json.dumps(data), headers=headers, verify=SSL_VERIFY)
    if ret.ok:
        return ret.json()['token']
    else:
        logger.error('error getting auth token: %s' % ret.content)
        raise Exception('Error getting auth token %s' % ret.content)

def forward_to_cyops(trigger, data, cyops_username, cyops_password):
    if data:
        full_uri = 'https://' + cs_host + '/api/triggers/1/' + trigger
        try:
            token = get_auth_token(cyops_username, cyops_password)
            headers = {
                'Authorization': 'Bearer {}'.format(token), "Content-Type": "application/json"
            }
            response_data = requests.post(full_uri, data, headers=headers, verify=SSL_VERIFY)
        except Exception as e:
            logging.exception("Error sending to CyOPs %s" % e)
            raise Exception("Error sending to CyOPs: %s" % e)

def get_key(host, port, username, source):
    return host.upper() + '-' + str(port) + '-' + username + '-' + source

class NotifierThread(Thread):
    def __init__(self, imap_config, username, cyops_username, cyops_password, trigger, thread_key, source):
        Thread.__init__(self)
        self.imap_config = imap_config
        self.username = username
        self.cyops_username = cyops_username
        self.cyops_password = cyops_password
        self.trigger = trigger
        self.thread_key = thread_key
        self.source = source
        self._create_imap_client()

    def _create_imap_client(self):
        get_auth_token(self.cyops_username, self.cyops_password)
        self.client = _make_imap_client(**self.imap_config)
        _select_folder(self.client, self.source)

    def run(self,**kwargs):
        try:
            global client_count
            idle_mode = False
            forward_to_cyops(self.trigger, "Fetching mail before Starting Push Notification Service for " + self.username,
                             self.cyops_username, self.cyops_password)
            while True:
                if not threads.get(self.thread_key, False):
                    break
                else:
                    if not threads[self.thread_key]['result']:
                        break
                if not idle_mode:
                    count = 0
                    try:
                        self.client.idle()
                        logger.info('Idle mode on')
                        idle_mode = True
                        count = 0
                    except Exception as e:
                        logger.exception(e)
                        count += 1
                        try:
                            self.client.logout()
                        except Exception as e:
                            logger.warn('IMAP Client logout error %s', str(e))
                        self._create_imap_client()
                        if(count>10):
                            raise Exception('Error in starting idle mode: %s' %e)

                if idle_mode:
                    # Wait for up to 30 seconds for an IDLE response
                    try:
                        responses = self.client.idle_check(idle_timeout)
                        if responses:
                            count = 0
                            command = responses[0][1].decode("utf-8")
                            if command == 'EXISTS':
                                logger.info('New mail received triggering: %s' %self.trigger)
                                forward_to_cyops(self.trigger, "New mail received triggering", self.cyops_username, self.cyops_password)
                            else:
                                if find_unread_mail(responses):
                                    logger.info('Mail marked as unread triggering: %s' % self.trigger)
                                    forward_to_cyops(self.trigger, "Mail marked as Unread", self.cyops_username, self.cyops_password)
                        else:
                            count += 1
                            total_time = (idle_timeout*count)/60
                            if total_time >= 20:
                                logger.info('Restarting Idle Mode after 20 min')
                                idle_mode = False
                                self.client.idle_done()
                                self.client.noop()
                    except Exception as e:
                        logger.error("Error in Listener %s" % e)
                        logger.info("Response from imap client %s" %responses)
            self.client.idle_done()
            self.client.noop()
            logger.info("logging out: %s" %self.username)
            client_count -= 1
            logout_client(self.client)
        except Exception as e:
            logger.error("Error in Thread %s" %e)

def start_server(host, port, ssl, username, password, source, destination, cyops_username, cyops_password, trigger):
    thread_key = get_key(host, port, username, source)
    if not threads.get(thread_key, False):
        threads[thread_key] = {"result": False, "message": "","threadobj": {}}
    if not threads[thread_key]['result']:
        try:
            imap_config = {'host': host, 'port': port, 'username': username, 'password': password, 'ssl': ssl}
            thread_instance = NotifierThread(imap_config, username, cyops_username, cyops_password, trigger, thread_key, source)
            thread_instance.start()
            threads[thread_key]['result'] = True
            threads[thread_key]['threadobj'] = thread_instance
            message = 'Server started for username: %s' % (username)
        except Exception as e:
            logger.info("returning the exception %s" % e)
            threads[thread_key]['message'] = str(e)
            return -1, str(e)
    else:
        message = 'Notification Service already running for server {0}:{1} and user {2} '.format(host,port,username)
        return -1, message
    return 0, message

def stop_server(username, port, host, source):
    thread_key = get_key(host, port, username, source)
    if threads.get(thread_key, False):
        threads.pop(thread_key)
        message = 'Notification Service stopped for username: %s' % (username)
    else:
        message = 'Notification Service already stopped for server {0}:{1} and user{2}'.format(host,port,username)
        logger.exception(message)
    return 0, message

def shutdown_server():
    global threads, client_count
    client_count = len(threads)
    threads = {}
    timeout = time.time() + 40
    while time.time() < timeout:
        if client_count <= 0:
            break
    return 0, "All Server Stopped."

def health_check(username, port, host, source):
    thread_key = get_key(host, port, username, source)
    if threads.get(thread_key, False):
        if threads[thread_key]['result']:
            if not threads[thread_key]['threadobj'].isAlive():
                message = 'Notifier Crashed for %s' % (username)
                logger.warn(message)
                return -1, message
            else:
                message = 'Found running notification for %s' % (username)
                return 0, message
        else:
            logger.error(threads[thread_key]['message'])
            return -1, threads[thread_key]['message']
    else:
        message = 'Notification Service not running for server {0}:{1} user {2}. ' \
                  'Deactivate and activate the connector to restart the listeners.'.format(host,port,username)
        logger.error(message)
        return -1, message

def handle(clientsocket):
    payload_bytes = clientsocket.recv(MAX_LENGTH)
    if payload_bytes:
        payload = payload_bytes.decode('utf-8')
        parser = argparse.ArgumentParser(description='Imap Mail Notification Actions')
        parser.add_argument('--start', help='Start Server', action='store_true', default=False, required=False)
        parser.add_argument('--stop', help='Stop Server', action='store_true', default=False, required=False)
        parser.add_argument('--exit', help='Stop Server', action='store_true', default=False, required=False)
        parser.add_argument('--check', help='Check Health', action='store_true', default=False, required=False)
        parser.add_argument('--host', help='Configuration for mail notification', required=False)
        parser.add_argument('--port', help='Port for server', required=False)
        parser.add_argument('--ssl', help='SSL for server', required=False)
        parser.add_argument('--username', help='Mail username to forward data', required=False)
        parser.add_argument('--password', help='Mail password to forward data', required=False)
        parser.add_argument('--source', help='Folder name to read mails', required=False)
        parser.add_argument('--destination', help='Folder name to keep read mails', required=False)
        parser.add_argument('--cyopsUsername', help='CyberSponse username to forward data', required=False)
        parser.add_argument('--cyopsPassword', help='CyberSponse password to forward data', required=False)
        parser.add_argument('--trigger', help='Configuration for trigger playbook', required=False)
        status = -1
        args_parsed = False
        try:
            # so that the main program does not exit if parsing fails
            args = parser.parse_args([p for p in re.split("( |\\\".*?\\\"|'.*?')", payload) if p.strip()])
            args_parsed = True
        except SystemExit as se:
            message = se
            logger.exception(se)
        if args_parsed:
            try:
                if args.exit:
                    status, message = shutdown_server()
                else:
                    if not args.port:
                        raise Exception('port is mandatory arguments')
                    if not args.host:
                        raise Exception('host is mandatory arguments')
                    if not args.username:
                        raise Exception('username is a mandatory argument')
                    if not args.source:
                        raise Exception('Source is a mandatory argument')
                    port = int(args.port)
                    host = args.host
                    ssl  = bool(args.ssl)
                    username = args.username
                    source = args.source
                    if args.stop:
                        status, message = stop_server(username, port, host, source)
                    elif args.check:
                        status, message = health_check(username, port, host, source)
                    else:
                        password = args.password
                        if not password:
                            raise Exception('password is a mandatory argument')
                        cyopsUsername = args.cyopsUsername
                        if not cyopsUsername:
                            raise Exception('CyOps Username is a mandatory argument')
                        cyopsPassword = args.cyopsPassword
                        if not cyopsPassword:
                            raise Exception('CyOps Password is a mandatory argument')
                        trigger = args.trigger
                        if not trigger:
                            raise Exception('Trigger is a mandatory argument')
                        destination = args.destination
                        if not destination:
                            raise Exception('Destination is a mandatory argument')
                        if args.start:
                            status, message = start_server(host, port, ssl, username, password, source,
                                                           destination, cyopsUsername, cyopsPassword, trigger)
                        else:
                            raise Exception('Unsupported function')
            except Exception as e:
                logger.exception(e)
                message = str(e)
        clientsocket.sendall(json.dumps({'status': status, 'message': message}).encode('utf-8'))
        clientsocket.close()
        if args.exit:
            serversocket.close()
            sys.exit()
# Bind socket to local host and port
try:
    serversocket.bind((HOST, PORT))
except socket.error as msg:
    logger.error('Bind failed: ' + msg)
    sys.exit()
serversocket.listen(10)

while True:
    (clientsocket, address) = serversocket.accept()
    handle(clientsocket)

# now keep talking with the client
